@@ -3,23 +3,28 @@
 # This Source Code Form is subject to the terms of the Mozilla Public
 # License, v. 2.0. If a copy of the MPL was not distributed with this
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
-from typing import Any, Awaitable, Callable, Dict, List, Optional, Set, Type, TypeVar, Union
+from __future__ import annotations
+
+from typing import Any, Awaitable, Callable, Type, TypeVar
+from uuid import UUID
 import asyncio
 
 from mautrix.util.logging import TraceLogger
 
-from .errors import UnexpectedError, UnexpectedResponse
+from .errors import AuthorizationFailedError, InternalError, RPCError, UnexpectedResponse
 from .rpc import CONNECT_EVENT, DISCONNECT_EVENT, SignaldRPCClient
 from .types import (
     Account,
     Address,
     Attachment,
     DeviceInfo,
+    ErrorMessage,
     GetIdentitiesResponse,
     Group,
     GroupID,
     GroupV2,
     IncomingMessage,
+    LinkPreview,
     LinkSession,
     Mention,
     Profile,
@@ -34,19 +39,20 @@ EventHandler = Callable[[T], Awaitable[None]]
 
 
 class SignaldClient(SignaldRPCClient):
-    _event_handlers: Dict[Type[T], List[EventHandler]]
-    _subscriptions: Set[str]
+    _event_handlers: dict[Type[T], list[EventHandler]]
+    _subscriptions: set[str]
 
     def __init__(
         self,
         socket_path: str = "/var/run/signald/signald.sock",
-        log: Optional[TraceLogger] = None,
-        loop: Optional[asyncio.AbstractEventLoop] = None,
+        log: TraceLogger | None = None,
+        loop: asyncio.AbstractEventLoop | None = None,
     ) -> None:
         super().__init__(socket_path, log, loop)
         self._event_handlers = {}
         self._subscriptions = set()
         self.add_rpc_handler("IncomingMessage", self._parse_message)
+        self.add_rpc_handler("ProtocolInvalidMessageError", self._parse_error)
         self.add_rpc_handler("WebSocketConnectionState", self._websocket_connection_state_change)
         self.add_rpc_handler("version", self._log_version)
         self.add_rpc_handler(CONNECT_EVENT, self._resubscribe)
@@ -70,7 +76,12 @@ class SignaldClient(SignaldRPCClient):
                 except Exception:
                     self.log.exception("Exception in event handler")
 
-    async def _parse_message(self, data: Dict[str, Any]) -> None:
+    async def _parse_error(self, data: dict[str, Any]) -> None:
+        if not data.get("error"):
+            return
+        await self._run_event_handler(ErrorMessage.deserialize(data))
+
+    async def _parse_message(self, data: dict[str, Any]) -> None:
         event_type = data["type"]
         event_data = data["data"]
         event_class = {
@@ -79,12 +90,12 @@ class SignaldClient(SignaldRPCClient):
         event = event_class.deserialize(event_data)
         await self._run_event_handler(event)
 
-    async def _log_version(self, data: Dict[str, Any]) -> None:
+    async def _log_version(self, data: dict[str, Any]) -> None:
         name = data["data"]["name"]
         version = data["data"]["version"]
         self.log.info(f"Connected to {name} v{version}")
 
-    async def _websocket_connection_state_change(self, change_event: Dict[str, Any]) -> None:
+    async def _websocket_connection_state_change(self, change_event: dict[str, Any]) -> None:
         evt = WebsocketConnectionStateChangeEvent.deserialize(
             {
                 "account": change_event["account"],
@@ -98,16 +109,12 @@ class SignaldClient(SignaldRPCClient):
             await self.request_v1("subscribe", account=username)
             self._subscriptions.add(username)
             return True
-        except UnexpectedError as e:
+        except RPCError as e:
             self.log.debug("Failed to subscribe to %s: %s", username, e)
-            evt = WebsocketConnectionStateChangeEvent(
-                state=(
-                    WebsocketConnectionState.AUTHENTICATION_FAILED
-                    if str(e) == "[401] Authorization failed!"
-                    else WebsocketConnectionState.DISCONNECTED
-                ),
-                account=username,
-            )
+            state = WebsocketConnectionState.DISCONNECTED
+            if isinstance(e, AuthorizationFailedError):
+                state = WebsocketConnectionState.AUTHENTICATION_FAILED
+            evt = WebsocketConnectionStateChangeEvent(state=state, account=username)
             await self._run_event_handler(evt)
             return False
 
@@ -116,11 +123,11 @@ class SignaldClient(SignaldRPCClient):
             await self.request_v1("unsubscribe", account=username)
             self._subscriptions.remove(username)
             return True
-        except UnexpectedError as e:
+        except RPCError as e:
             self.log.debug("Failed to unsubscribe from %s: %s", username, e)
             return False
 
-    async def _resubscribe(self, unused_data: Dict[str, Any]) -> None:
+    async def _resubscribe(self, unused_data: dict[str, Any]) -> None:
         if self._subscriptions:
             self.log.debug("Resubscribing to users")
             for username in list(self._subscriptions):
@@ -137,9 +144,7 @@ class SignaldClient(SignaldRPCClient):
                 )
                 await self._run_event_handler(evt)
 
-    async def register(
-        self, phone: str, voice: bool = False, captcha: Optional[str] = None
-    ) -> str:
+    async def register(self, phone: str, voice: bool = False, captcha: str | None = None) -> str:
         resp = await self.request_v1("register", account=phone, voice=voice, captcha=captcha)
         return resp["account_id"]
 
@@ -150,6 +155,9 @@ class SignaldClient(SignaldRPCClient):
     async def start_link(self) -> LinkSession:
         return LinkSession.deserialize(await self.request_v1("generate_linking_uri"))
 
+    async def wait_for_scan(self, session_id: str) -> None:
+        await self.request_v1("wait_for_scan", session_id=session_id)
+
     async def finish_link(
         self, session_id: str, device_name: str = "mausignald", overwrite: bool = False
     ) -> Account:
@@ -160,8 +168,8 @@ class SignaldClient(SignaldRPCClient):
 
     @staticmethod
     def _recipient_to_args(
-        recipient: Union[Address, GroupID], simple_name: bool = False
-    ) -> Dict[str, Any]:
+        recipient: Address | GroupID, simple_name: bool = False
+    ) -> dict[str, Any]:
         if isinstance(recipient, Address):
             recipient = recipient.serialize()
             field_name = "address" if simple_name else "recipientAddress"
@@ -170,17 +178,22 @@ class SignaldClient(SignaldRPCClient):
         return {field_name: recipient}
 
     async def react(
-        self, username: str, recipient: Union[Address, GroupID], reaction: Reaction
+        self,
+        username: str,
+        recipient: Address | GroupID,
+        reaction: Reaction,
+        req_id: UUID | None = None,
     ) -> None:
         await self.request_v1(
             "react",
             username=username,
             reaction=reaction.serialize(),
+            req_id=req_id,
             **self._recipient_to_args(recipient),
         )
 
     async def remote_delete(
-        self, username: str, recipient: Union[Address, GroupID], timestamp: int
+        self, username: str, recipient: Address | GroupID, timestamp: int
     ) -> None:
         await self.request_v1(
             "remote_delete",
@@ -192,16 +205,19 @@ class SignaldClient(SignaldRPCClient):
     async def send(
         self,
         username: str,
-        recipient: Union[Address, GroupID],
+        recipient: Address | GroupID,
         body: str,
-        quote: Optional[Quote] = None,
-        attachments: Optional[List[Attachment]] = None,
-        mentions: Optional[List[Mention]] = None,
-        timestamp: Optional[int] = None,
+        quote: Quote | None = None,
+        attachments: list[Attachment] | None = None,
+        mentions: list[Mention] | None = None,
+        previews: list[LinkPreview] | None = None,
+        timestamp: int | None = None,
+        req_id: UUID | None = None,
     ) -> None:
         serialized_quote = quote.serialize() if quote else None
         serialized_attachments = [attachment.serialize() for attachment in (attachments or [])]
         serialized_mentions = [mention.serialize() for mention in (mentions or [])]
+        serialized_previews = [preview.serialize() for preview in (previews or [])]
         resp = await self.request_v1(
             "send",
             username=username,
@@ -209,7 +225,9 @@ class SignaldClient(SignaldRPCClient):
             attachments=serialized_attachments,
             quote=serialized_quote,
             mentions=serialized_mentions,
+            previews=serialized_previews,
             timestamp=timestamp,
+            req_id=req_id,
             **self._recipient_to_args(recipient),
         )
 
@@ -267,8 +285,8 @@ class SignaldClient(SignaldRPCClient):
         self,
         username: str,
         sender: Address,
-        timestamps: List[int],
-        when: Optional[int] = None,
+        timestamps: list[int],
+        when: int | None = None,
         read: bool = False,
     ) -> None:
         if not read:
@@ -278,25 +296,25 @@ class SignaldClient(SignaldRPCClient):
             "mark_read", account=username, timestamps=timestamps, when=when, to=sender.serialize()
         )
 
-    async def list_accounts(self) -> List[Account]:
+    async def list_accounts(self) -> list[Account]:
         resp = await self.request_v1("list_accounts")
         return [Account.deserialize(acc) for acc in resp.get("accounts", [])]
 
     async def delete_account(self, username: str, server: bool = False) -> None:
         await self.request_v1("delete_account", account=username, server=server)
 
-    async def get_linked_devices(self, username: str) -> List[DeviceInfo]:
+    async def get_linked_devices(self, username: str) -> list[DeviceInfo]:
         resp = await self.request_v1("get_linked_devices", account=username)
         return [DeviceInfo.deserialize(dev) for dev in resp.get("devices", [])]
 
     async def remove_linked_device(self, username: str, device_id: int) -> None:
         await self.request_v1("remove_linked_device", account=username, deviceId=device_id)
 
-    async def list_contacts(self, username: str) -> List[Profile]:
+    async def list_contacts(self, username: str) -> list[Profile]:
         resp = await self.request_v1("list_contacts", account=username)
         return [Profile.deserialize(contact) for contact in resp["profiles"]]
 
-    async def list_groups(self, username: str) -> List[Union[Group, GroupV2]]:
+    async def list_groups(self, username: str) -> list[Group | GroupV2]:
         resp = await self.request_v1("list_groups", account=username)
         legacy = [Group.deserialize(group) for group in resp.get("legacyGroups", [])]
         v2 = [GroupV2.deserialize(group) for group in resp.get("groups", [])]
@@ -306,11 +324,11 @@ class SignaldClient(SignaldRPCClient):
         self,
         username: str,
         group_id: GroupID,
-        title: Optional[str] = None,
-        avatar_path: Optional[str] = None,
-        add_members: Optional[List[Address]] = None,
-        remove_members: Optional[List[Address]] = None,
-    ) -> Union[Group, GroupV2, None]:
+        title: str | None = None,
+        avatar_path: str | None = None,
+        add_members: list[Address] | None = None,
+        remove_members: list[Address] | None = None,
+    ) -> Group | GroupV2 | None:
         update_params = {
             key: value
             for key, value in {
@@ -338,7 +356,7 @@ class SignaldClient(SignaldRPCClient):
 
     async def get_group(
         self, username: str, group_id: GroupID, revision: int = -1
-    ) -> Optional[GroupV2]:
+    ) -> GroupV2 | None:
         resp = await self.request_v1(
             "get_group", account=username, groupID=group_id, revision=revision
         )
@@ -348,7 +366,7 @@ class SignaldClient(SignaldRPCClient):
 
     async def get_profile(
         self, username: str, address: Address, use_cache: bool = False
-    ) -> Optional[Profile]:
+    ) -> Profile | None:
         try:
             # async is a reserved keyword, so can't pass it as a normal parameter
             kwargs = {"async": use_cache}
@@ -368,7 +386,7 @@ class SignaldClient(SignaldRPCClient):
         return GetIdentitiesResponse.deserialize(resp)
 
     async def set_profile(
-        self, username: str, name: Optional[str] = None, avatar_path: Optional[str] = None
+        self, username: str, name: str | None = None, avatar_path: str | None = None
     ) -> None:
         args = {}
         if name is not None:
@@ -382,8 +400,8 @@ class SignaldClient(SignaldRPCClient):
         username: str,
         recipient: Address,
         trust_level: str,
-        safety_number: Optional[str] = None,
-        qr_code_data: Optional[str] = None,
+        safety_number: str | None = None,
+        qr_code_data: str | None = None,
     ) -> None:
         args = {}
         if safety_number:
@@ -401,3 +419,9 @@ class SignaldClient(SignaldRPCClient):
             trust_level=trust_level,
             address=recipient.serialize(),
         )
+
+    async def find_uuid(self, username: str, number: str) -> UUID | None:
+        resp = await self.request_v1(
+            "resolve_address", partial=Address(number=number).serialize(), account=username
+        )
+        return Address.deserialize(resp).uuid
