@@ -163,6 +163,7 @@ class Group(SerializableAttrs):
 class GroupV2ID(SerializableAttrs):
     id: GroupID
     revision: Optional[int] = None
+    removed: Optional[bool] = False
 
 
 class AccessControlMode(SerializableEnum):
@@ -275,7 +276,7 @@ class Sticker(SerializableAttrs):
 
 @dataclass
 class RemoteDelete(SerializableAttrs):
-    target_sent_timestamp: int = field(json="targetSentTimestamp")
+    target_sent_timestamp: int
 
 
 class SharedContactDetailType(SerializableEnum):
@@ -346,6 +347,14 @@ class SharedContact(SerializableAttrs):
     address: Optional[SharedContactAddress] = None
 
 
+@dataclass
+class LinkPreview(SerializableAttrs):
+    url: str
+    title: str
+    description: str
+    attachment: Optional[Attachment] = None
+
+
 @dataclass
 class MessageData(SerializableAttrs):
     timestamp: int
@@ -368,6 +377,8 @@ class MessageData(SerializableAttrs):
 
     remote_delete: Optional[RemoteDelete] = field(default=None, json="remoteDelete")
 
+    previews: List[LinkPreview] = field(factory=lambda: [])
+
     @property
     def is_message(self) -> bool:
         return bool(self.body or self.attachments or self.sticker or self.contacts)
@@ -529,6 +540,24 @@ class IncomingMessage(SerializableAttrs):
     receipt_message: Optional[ReceiptMessage] = None
 
 
+@dataclass(kw_only=True)
+class ErrorMessageData(SerializableAttrs):
+    sender: str
+    timestamp: int
+    message: str
+    sender_device: int
+    content_hint: int
+
+
+@dataclass(kw_only=True)
+class ErrorMessage(SerializableAttrs):
+    type: str
+    version: str
+    data: ErrorMessageData
+    error: bool
+    account: str
+
+
 class WebsocketConnectionState(SerializableEnum):
     # States from signald itself
     DISCONNECTED = "DISCONNECTED"
