@@ -4,11 +4,11 @@ from mautrix.bridge.commands import CommandEvent as BaseCommandEvent
 
 if TYPE_CHECKING:
     from ..__main__ import SignalBridge
-    from ..user import User
     from ..portal import Portal
+    from ..user import User
 
 
 class CommandEvent(BaseCommandEvent):
-    bridge: 'SignalBridge'
-    sender: 'User'
-    portal: 'Portal'
+    bridge: "SignalBridge"
+    sender: "User"
+    portal: "Portal"
