@@ -128,16 +128,21 @@ class MatrixHandler(BaseMatrixHandler):
         event_id: EventID,
         data: SingleReceiptEventContent,
     ) -> None:
-        await portal.handle_read_receipt(event_id, data)
-
         message = await DBMessage.get_by_mxid(event_id, portal.mxid)
         if not message:
             return
 
         user.log.trace(f"Sending read receipt for {message.timestamp} to {message.sender}")
-        await self.signal.send_receipt(
-            user.username, message.sender, timestamps=[message.timestamp], when=data.ts, read=True
-        )
+        try:
+            await self.signal.send_receipt(
+                user.username,
+                message.sender,
+                timestamps=[message.timestamp],
+                when=data.ts,
+                read=True,
+            )
+        except Exception as e:
+            await user.handle_auth_failure(e)
 
     async def handle_typing(self, room_id: RoomID, typing: list[UserID]) -> None:
         pass
